/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<vcg/complex/algorithms/update/color.h>
#include<vcg/complex/algorithms/point_sampling.h>
#include<vcg/complex/algorithms/clean.h>
#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/import_obj.h>

#include<wrap/io_trimesh/export_off.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/complex/algorithms/clean.h>

using namespace vcg;

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Color4b, vcg::vertex::Qualityf, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::FFAdj,  vcg::face::VertexRef, vcg::face::Color4b, vcg::face::Normal3f, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<   MyUsedTypes, vcg::edge::VertexRef, vcg::edge::EEAdj, vcg::edge::BitFlags > {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};


template <class MeshType>
void UniformCreaseEdgeSampling(MeshType &m, // the mesh that has to be sampled
                     std::vector<typename MeshType::CoordType> &UniformEdgeSamples, // the vector that will contain the set of points sample on the crease edges (including corners)
                     typename MeshType::ScalarType &radius,  // the Poisson Disk Radius (used if sampleNum==0, setted if sampleNum!=0)
                     float creaseAngleDeg // the angle that defines the crease edges
)
{
// we search for all the edges where the difference between the normals of the two incident faces 
 // is greater than a given threshold

  MeshType edgeMesh; // will contains all the crease edges
  float cosThr = cosf( math::ToRad(creaseAngleDeg) );
  for(auto fi = m.face.begin(); fi != m.face.end(); ++fi)
  {
    auto &f = *fi;
    for(int i = 0; i < 3; i++)
    {
      float dot = f.N() * f.FFp(i)->N();
      if((dot < cosThr) && (f.FFp(i) > &f) )
        tri::Allocator<MyMesh>::AddEdge(edgeMesh, f.P0(i), f.P1(i));
    }
  }

  tri::UpdateBounding<MyMesh>::Box(edgeMesh); 
  tri::Clean<MyMesh>::RemoveDuplicateVertex(edgeMesh);
  tri::Allocator<MyMesh>::CompactEveryVector(edgeMesh);
  tri::Clean<MyMesh>::SelectNonManifoldVertexOnEdgeMesh(edgeMesh);
  tri::Clean<MyMesh>::SplitSelectedVertexOnEdgeMesh(edgeMesh);
  tri::Allocator<MyMesh>::CompactEveryVector(edgeMesh);
  
  tri::TrivialSampler<MyMesh> ts(UniformEdgeSamples);  
  tri::SurfaceSampling<MyMesh, tri::TrivialSampler < MyMesh> >::EdgeMeshUniform(edgeMesh, ts, radius); 

  printf("Sampled %i points\n",ts.SampleVec().size());
}

template <class MeshType>
void CreaseAwarePoissonDiskSampling(MeshType &m, // the mesh that has to be sampled
                     std::vector<typename MeshType::CoordType> &poissonSamples, // the vector that will contain the set of points
                     typename MeshType::ScalarType &radius,  // the Poisson Disk Radius (used if sampleNum==0, setted if sampleNum!=0)
                    float creaseAngleDeg, // the angle that defines the crease edges
                     unsigned int randSeed=0)

{
  typedef tri::TrivialSampler<MeshType> BaseSampler;
  typedef tri::MeshSampler<MeshType> MontecarloSampler;
  typename tri::SurfaceSampling<MeshType, BaseSampler>::PoissonDiskParam pp;

  int t0=clock();
  std::vector<MyMesh::CoordType> samples;
  UniformCreaseEdgeSampling(m, samples, radius, creaseAngleDeg);
  MyMesh meshCreaseEdgeSamples;
  tri::BuildMeshFromCoordVector(meshCreaseEdgeSamples,samples);
 
  MyMesh MontecarloMesh; // it will contains the samples generated by the montecarlo sampling over the input mesh 
  tri::MeshSampler<MyMesh> mcSampler(MontecarloMesh);
  tri::SurfaceSampling<MyMesh, tri::MeshSampler<MyMesh> >::Montecarlo(m, mcSampler, 50000);
  tri::UpdateBounding<MyMesh>::Box(MontecarloMesh);

  tri::SurfaceSampling<MyMesh, tri::MeshSampler < MyMesh> >::PoissonDiskParam pp;
  pp.preGenMesh = &meshCreaseEdgeSamples;
  pp.preGenFlag = true;
  pp.bestSamplePoolSize = 20;
  
  tri::TrivialSampler<MyMesh> pdSampler(poissonSamples);   
  tri::SurfaceSampling<MyMesh, tri::MeshSampler < MyMesh> >::PoissonDiskPruning(pdSampler, MontecarloMesh, radius, pp);
}




int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.obj>\n");
    return -1;
  }
  /*!
    */
  MyMesh m;
  srand(time(NULL));
  float samplingRadius = 0.02f;
  if(vcg::tri::io::ImporterOFF<MyMesh>::Open(m,argv[1])!=vcg::tri::io::ImporterOFF<MyMesh>::NoError)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  if(argc>2)
  {
   samplingRadius=atof(argv[2]);
  }

  float radius = m.bbox.Diag() * samplingRadius;
  printf("Sampling radius %f\n",radius);

  tri::RequirePerVertexNormal(m);
  tri::RequirePerFaceNormal(m);
  tri::UpdateNormal<MyMesh>::PerVertexNormalized(m);
  tri::UpdateNormal<MyMesh>::PerFaceNormalized(m);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateColor<MyMesh>::PerVertexConstant(m,vcg::Color4b::White);
  tri::UpdateQuality<MyMesh>::VertexConstant(m,0.0f);

  
  vcg::tri::io::ExporterPLY<MyMesh>::Save(poissonMesh,"PoissonMesh.ply");

  return 0;
}
